import functools

import cherrypy
import ujson

from cherryontop.errors import HTTPError, InvalidParameter, MissingParameter
from cherryontop.errors import PayloadError, RequestError, UnexpectedParameter


def validate_body(*a, **kw):
    allowed, required, validators = _get_checks(*a, **kw)

    def wrap(f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            body = _deserialize_request_body()
            param_names = body.keys()

            # 1. all required parameters present?
            for param in required:
                if param not in param_names:
                    raise MissingParameter(param)

            # 2. all supplied parameters allowed?
            for param in param_names:
                if param not in allowed:
                    raise UnexpectedParameter(param)

            # 3. parameter values pass all tests?
            for param_name, test in validators:
                if param_name in param_names:
                    _validate_param(test, param_name, body[param_name])

            # if a body param and a query param match in name, only the body
            # param will be passed through
            kwargs.update(**body)

            return f(*args, **kwargs)
        return wrapped
    return wrap


def _get_checks(*args, **kwargs):
    required = kwargs.pop('require', [])
    required = set(required)

    allowed = kwargs.pop('allow', [])
    allowed = set(allowed)
    allowed |= required

    to_check = []

    for validator in args:
        param_name, tests = validator
        if isinstance(tests, (list, tuple,)):
            for test in tests:
                _check_validator(test)
                to_check.append((param_name, test,))
        else:
            _check_validator(tests)
            to_check.append(validator)

        allowed.add(param_name)

    return allowed, required, to_check


def _check_validator(test):
    if not isinstance(test, type) and not callable(test):
        raise TypeError('validator must be callable')


def _deserialize_request_body():
    mimetype = cherrypy.request.headers.get('content-type')
    if mimetype != 'application/json':
        raise RequestError('Not JSON')

    body = cherrypy.request.body.read()
    if not body:
        return {}

    try:
        body = ujson.loads(body)
    except ValueError:
        raise PayloadError('Invalid JSON')

    return body


def _validate_param(test, param_name, value):
    if isinstance(test, type):
        if not isinstance(value, test):
            raise InvalidParameter(param_name)
    elif callable(test):
        if not test(value):
            raise InvalidParameter(param_name)



import cherrypy
import ujson

from cherryontop.errors.base import CherryOnTopError, HTTPError
from cherryontop.errors.req import InvalidParameter, MissingParameter
from cherryontop.errors.req import PayloadError, RequestError
from cherryontop.errors.req import UnexpectedParameter


def _body(error, http_response_code):
    return ujson.dumps({
        'error': error.__class__.__name__,
        'http_response_code': http_response_code,
        'message': error.message,
    })


def _extra_headers(status):
    if status == 401:
        realm = 'Basic realm="auth required"'
        cherrypy.response.headers['www-authenticate'] = realm


def error_response_handler():
    _, e, _ = cherrypy._cperror._exc_info()
    status = e.http_response_code if isinstance(e, CherryOnTopError) else 500

    cherrypy.response.body = _body(e, status)
    cherrypy.response.headers['content-type'] = 'application/json'
    cherrypy.response.status = status
    _extra_headers(status)


__all__ = [
    'error_response_handler',
    'CherryOnTopError', 'HTTPError', 'InvalidParameter', 'MissingParameter',
    'PayloadError', 'RequestError', 'UnexpectedParameter',
]



from cherryontop.decorators.body import validate_body
from cherryontop.decorators.qp import typecast_query_params
from cherryontop.decorators.routes import delete, get, post, put, route


__all__ = [
    'delete', 'get', 'post', 'put', 'route',
    'typecast_query_params', 'validate_body',
]



from cherryontop.cache.controller import icontrollers, register_controller
from cherryontop.cache.route import map_routes_for_controller, register_route


def map_all_routes(dispatcher):
    for cid, controller in icontrollers():
        map_routes_for_controller(cid, controller, dispatcher)


__all__ = ['map_all_routes', 'register_controller', 'register_route',]



from cherryontop.errors.base import HTTPError


class RequestError(HTTPError):

    http_response_code = 400


class PayloadError(RequestError):
    pass


class InvalidParameter(PayloadError):
    pass


class MissingParameter(PayloadError):
    pass


class UnexpectedParameter(PayloadError):
    pass



"""
3 - Dynamic routes

Dynamic routes in CherryOnTop may be specified by placing a colon
prior to each dynamic component, then passing those elements into
the method as positional or keyword arguments.

"""

import cherryontop


class HelloWorld(cherryontop.Controller):

    @cherryontop.get('/hello')
    @cherryontop.get('/hello/:name')
    def hello_world(self, name='world'):
        return {'message': 'Hello %s!' % name}

    @cherryontop.get('/goodbye/:name')
    def goodbye(self, name):
        return {'message': 'So long %s!' % name}


if __name__ == '__main__':
    cherryontop.start_server(log_to_screen=True)



import cherrypy

from cherryontop.cache import map_all_routes


def _create_dispatcher():
    dispatcher = cherrypy.dispatch.RoutesDispatcher()
    map_all_routes(dispatcher)
    return dispatcher


def _daemonize():
    daemonizer = cherrypy.process.plugins.Daemonizer(cherrypy.engine)
    daemonizer.subscribe()
    sig_handler = cherrypy.process.plugins.SignalHandler(cherrypy.engine)
    sig_handler.subscribe()


def _set_default_headers():
    headers = cherrypy.config.get('tools.response_headers.headers', [])
    headers.append(('Content-Type', 'application/json'))

    cherrypy.config.update({
        'tools.response_headers.headers': headers,
        'tools.response_headers.on': True,
    })


def _setup_global_config(config=None):
    if config:  # dictionary or path to config file
        cherrypy.config.update(config)
    _set_default_headers()


def start_server(host='0.0.0.0',
                 port=8080,
                 threads=20,
                 autoreload=False,
                 log_to_screen=False,
                 path_to_access_log=None,
                 path_to_error_log=None,
                 **kwargs):
    conf = {
        'engine.autoreload_on': autoreload,
        'log.screen': log_to_screen,
        'server.socket_host': host,
        'server.socket_port': port,
        'server.thread_pool': threads,
    }

    if path_to_access_log:
        conf['log.access_file'] = path_to_access_log
    if path_to_error_log:
        conf['log.error_file'] = path_to_error_log

    strict_start_server(config=conf, **kwargs)


def strict_start_server(config=None, daemonize=False):
    _setup_global_config(config)

    if daemonize:
        _daemonize()

    app_config = {
        '/': {'request.dispatch': _create_dispatcher()},
    }

    cherrypy.quickstart(None, '/', config=app_config)



from cherryontop.cache import map_all_routes
from cherryontop.controller import Controller
from cherryontop.decorators import delete, get, post, put, route
from cherryontop.decorators import typecast_query_params, validate_body
from cherryontop.spinup import start_server, strict_start_server


__all__ = [
    'Controller', 'delete', 'get', 'map_all_routes', 'post', 'put', 'route',
    'start_server', 'strict_start_server',
    'typecast_query_params', 'validate_body',
]



"""
6 - Request bodies

For handlers expecting JSON dictionary payloads, CherryOnTop
provides the `validate_body` decorator to extract parameters,
validate them, and stuff them into the method as keyword arguments.

It takes any number of (param_name, validator) tuples, where
`param_name` is the key and `validator` may be a callable or
type. Callables must take the parameter value as an input and
return True/False (a False immediately triggers an `InvalidParameter`
error), and types indicate we should test whether the given value
is an instance of the specified type (`InvalidParameter` also thrown
if not).

Like `typecast_query_params` all expected parameters must be
declared, else `UnexpectedParameter` is thrown. Use the `allow`
keyword argument to bypass this.

The new keyword argument `require` is available here too. Any
parameter listed in the array but not provided in the request
body will trigger a `MissingParameter` error.

NB: Injecting both body parameters and query parameters as
keyword arguments leads to the possibility for same-named
parameters to overwrite one another. Hence, it is recommended
you avoid naming conflicts across body and query parameters
within a single handler.

Caveat: since cherrypy passes query parameters as keyword
arguments by default, `typecast_query_params` must be listed
above `validate_body` any time they are used together.

$ curl 0.0.0.0:8080 -H 'content-type:application/json' -d '{"name":"chris"}'
{"message":"Hello chris!"}

$ curl 0.0.0.0:8080 -H 'content-type:application/json' -H "content-length:0" -X POST
{"message":"name","http_response_code":400,"error":"MissingParameter"}

$ curl 0.0.0.0:8080 -H 'content-type:application/json' -d '{"name":"chris","underwear_size":"m","foo":"bar"}'
{"message":"Hello chris!","foo":"bar","underwear_size":"m"}

$ curl 0.0.0.0:8080 -H 'content-type:application/json' -d '{"name":"chris","underwear_size":"M"}'
{"message":"underwear_size","http_response_code":400,"error":"InvalidParameter"}

"""

import cherryontop


UNDERWEAR_SIZES = ['s', 'm', 'l']


class HelloWorld(cherryontop.Controller):

    @cherryontop.post('/')
    @cherryontop.validate_body(('name', unicode,),
                               ('underwear_size', lambda i: i in UNDERWEAR_SIZES,),
                               allow=('foo',),
                               require=('name',))
    def hello_world(self, name=None, **kw):
        kw['message'] = 'Hello %s!' % (name or 'you')
        return kw


if __name__ == '__main__':
    cherryontop.start_server(log_to_screen=True)



"""
1 - Hello World

This first sample app defines a single handler, opens a GET route
at '/', and starts up a server with screen logging enabled.

$ python tutorial01.py
$ curl 0.0.0.0:8080 -v

* About to connect() to 0.0.0.0 port 8080 (#0)
*   Trying 0.0.0.0...
* Adding handle: conn: 0x7f97d0804000
* Adding handle: send: 0
* Adding handle: recv: 0
* Curl_addHandleToPipeline: length: 1
* - Conn 0 (0x7f97d0804000) send_pipe: 1, recv_pipe: 0
* Connected to 0.0.0.0 (0.0.0.0) port 8080 (#0)
> GET / HTTP/1.1
> User-Agent: curl/7.30.0
> Host: 0.0.0.0:8080
> Accept: */*
>
< HTTP/1.1 200 OK
< Date: Fri, 07 Nov 2014 23:56:33 GMT
< Content-Length: 26
< Content-Type: application/json
* Server CherryPy/3.6.0 is not blacklisted
< Server: CherryPy/3.6.0
<
* Connection #0 to host 0.0.0.0 left intact
{"message":"Hello world!"}

"""

import cherryontop


class HelloWorld(cherryontop.Controller):

    @cherryontop.get('/')
    def hello_world(self):
        return {'message': 'Hello world!'}


if __name__ == '__main__':
    cherryontop.start_server(log_to_screen=True)



"""
2 - Additional routes

Any number of routes may be bound by defining additional handlers
wrapped with the `route`, `get`, `post`, `put`, or `delete`
decorators.

`route` is a special case only in that it provides support for
multiple methods.

It is worth mentioning here that it is completely safe to wrap the
same handler with multiple route decorators, e.g. `post_hello_world`
below.

Sample app 2 supports the following routes:
    GET     /
    POST    /
    POST    /other_post
    PUT     /
    DELETE  /
    GET     /other_get

"""

import cherryontop
import cherrypy


class HelloWorld(cherryontop.Controller):

    @staticmethod
    def _render(handler_name):
        return {
            'message': 'Hello world!',
            'method': cherrypy.request.method,
            'handler': handler_name,
        }

    @cherryontop.get('/')
    def get_hello_world(self):
        return self._render('get_hello_world')

    @cherryontop.post('/')
    @cherryontop.post('/other_post')
    def post_hello_world(self):
        return self._render('post_hello_world')

    @cherryontop.route('/', methods=['PUT', 'DELETE'])
    @cherryontop.get('/other_get')
    def other_hello_world(self):
        return self._render('other_hello_world')


if __name__ == '__main__':
    cherryontop.start_server(log_to_screen=True)



import functools
import uuid

import ujson

from cherryontop.cache import register_controller, register_route
from cherryontop.errors import error_response_handler


def _to_json(f):
    @functools.wraps(f)
    def wrap(*a, **kw):
        data = f(*a, **kw)
        return ujson.dumps(data)
    return wrap


def _setup_cp_config(cls):
    cp_config = getattr(cls, '_cp_config', {})
    if 'request.error_response' not in cp_config:
        cp_config['request.error_response'] = error_response_handler
        cls._cp_config = cp_config


class Base(type):
    def __new__(cls, name, bases, dct):
        bases_of_this_type = [b for b in bases if isinstance(b, cls)]
        is_proper_subclass = len(bases_of_this_type) > 0

        if is_proper_subclass:  # skip for the base class
            cid = uuid.uuid4().hex

            for key, val in dct.items():
                if hasattr(val, '_cbs_routes'):  # handler?
                    for route, method in val._cbs_routes:
                        register_route(cid,
                                       route,
                                       val.__name__,
                                       conditions={'method': [method]})

                    dct[key] = _to_json(val)

        the_class = super(Base, cls).__new__(cls, name, bases, dct)

        if is_proper_subclass:
            register_controller(cid, the_class)
            _setup_cp_config(the_class)

        return the_class


class Controller(object):

    __metaclass__ = Base



from collections import defaultdict
import uuid


_registry = defaultdict(list)


def register_route(cid, route, handler_name, conditions=None):
    _registry[cid].append(dict(
        route=route,
        action=handler_name,
        conditions=conditions,
        name=uuid.uuid4().hex,
    ))


def map_routes_for_controller(cid, controller, dispatcher):
    for kwargs in _registry[cid]:
        dispatcher.connect(controller=controller, **kwargs)



"""
4 - Exceptions

CherryOnTop exception classes plug into the cherrypy engine to
return JSON error descriptions to the client. That is, when a
subclass of `CherryOnTopError` is raised e.g.

    raise cherryontop.errors.HTTPError('something bad happened')

and not explicitly caught, the following cascade kicks in:

  - the HTTP response code is set to <exception>.http_response_code
  - the HTTP response type is set to 'application/json'
  - a JSON payload is rendered and returned, e.g.

    {
        'error': 'HTTPError',
        'http_response_code': 500,
        'message': 'something bad happened',
    }

$ curl "0.0.0.0:8080/broken" -v

* About to connect() to 0.0.0.0 port 8080 (#0)
*   Trying 0.0.0.0...
* Adding handle: conn: 0x7fc573804000
* Adding handle: send: 0
* Adding handle: recv: 0
* Curl_addHandleToPipeline: length: 1
* - Conn 0 (0x7fc573804000) send_pipe: 1, recv_pipe: 0
* Connected to 0.0.0.0 (0.0.0.0) port 8080 (#0)
> GET /broken HTTP/1.1
> User-Agent: curl/7.30.0
> Host: 0.0.0.0:8080
> Accept: */*
>
< HTTP/1.1 500 Internal Server Error
< Date: Sun, 09 Nov 2014 14:48:27 GMT
< Content-Length: 81
< Content-Type: application/json
* Server CherryPy/3.6.0 is not blacklisted
< Server: CherryPy/3.6.0
<
* Connection #0 to host 0.0.0.0 left intact
{"message":"this handler is broken","http_response_code":500,"error":"HTTPError"}

"""

import cherryontop


class HelloWorld(cherryontop.Controller):

    @cherryontop.get('/broken')
    def im_broken(self):
        raise cherryontop.errors.HTTPError('this handler is broken')


if __name__ == '__main__':
    cherryontop.start_server(log_to_screen=True)



"""
5 - Query parameters

While standard cherrypy handling for query parameters is still
intact, CherryOnTop's `typecast_query_params` decorator allows
you to define a type signature to cast them on the way in. It
takes any number of (param, func) tuples, where `param` is the
name of the query parameter and `func` is a callable that
casts the given value to the desired type.

`UnexpectedParameter` is raised when any undeclared parameter is
encountered. Use the `allow` keyword argument to bypass this
behavior for parameters that need no typecasting. When
`UnexpectedParameter` is raised, the name of the offending
parameter is returned in the response payload's `message` key.

$ curl '0.0.0.0:8080'
{"message":"Hello you!"}

$ curl '0.0.0.0:8080?name=chris'
{"message":"Hello chris!"}

$ curl '0.0.0.0:8080?name=chris&a=1&b=2'
{"a":1,"message":"Hello chris!","b":2.0}

$ curl '0.0.0.0:8080?name=chris&a=1&b=2&c=3'
{"message":"c","http_response_code":400,"error":"UnexpectedParameter"}

"""

import cherryontop


class HelloWorld(cherryontop.Controller):

    @cherryontop.get('/')
    @cherryontop.typecast_query_params(('a', int,),
                                       ('b', float,),
                                       allow=['name'])
    def hello_world(self, name=None, **kw):
        kw['message'] = 'Hello %s!' % (name or 'you')
        return kw


if __name__ == '__main__':
    cherryontop.start_server(log_to_screen=True)



import functools

from cherryontop.errors import InvalidParameter, UnexpectedParameter


def typecast_query_params(*a, **kw):
    allowed, cast_funcs = _get_checks(*a, **kw)

    def wrap(f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            # all supplied parameters allowed?
            for param in kwargs:
                if param not in allowed:
                    raise UnexpectedParameter(param)

            # typecast params
            for param_name, cast in cast_funcs:
                if param_name in kwargs:
                    try:
                        kwargs[param_name] = cast(kwargs[param_name])
                    except ValueError:
                        raise InvalidParameter(param_name)

            return f(*args, **kwargs)
        return wrapped
    return wrap


def _get_checks(*args, **kwargs):
    allowed = kwargs.pop('allow', [])
    allowed = set(allowed)

    to_cast = []

    for caster in args:
        param_name, func = caster
        if not callable(func):
            raise TypeError('cast func must be callable')

        allowed.add(param_name)
        to_cast.append(caster)

    return allowed, to_cast



"""
7 - Starting the server

As you've seen in the tutorials so far, it's trivial to spin
up CherryOnTop apps by calling `start_server`. This tutorial
illustrates additional options you may pass to `start_server`,
as well as the preferred way to pass cherrypy-style configs.

"""

import os

import cherryontop


class HelloWorld(cherryontop.Controller):

    @cherryontop.get('/')
    def hello_world(self):
        return {'message': 'Hello world!'}


if __name__ == '__main__':
    file_dir = os.path.dirname(os.path.realpath(__file__))
    access_log = '%s/access.log' % file_dir
    error_log = '%s/error.log' % file_dir


    # quick and easy - no config file but options for
    # common parameters

    cherryontop.start_server(host='127.0.0.1',    # default: 0.0.0.0
                             port=8123,           # default: 8080
                             threads=10,          # default: 20
                             autoreload=True,     # default: False
                             log_to_screen=True,  # default: False
                             path_to_access_log=access_log,
                             path_to_error_log=error_log)


    # an equivalent example - how to drive servers using
    # cherrypy config files

    config_file = '%s/tutorial.conf' % file_dir
    cherryontop.strict_start_server(config_file)


    # equivalently, a cherrypy config dict

    config_data = {
        'engine.autoreload_on': True,
        'log.access_file': access_log,
        'log.error_file': error_log,
        'log.screen': True,
        'server.socket_host': '127.0.0.1',
        'server.socket_port': 8123,
        'server.thread_pool': 10,
    }
    cherryontop.strict_start_server(config_data)



from setuptools import find_packages, setup


long_desc = '''\
CherryOnTop is a thin layer atop CherryPy that provides convenience methods
for binding routes and handling JSON request/response payloads.
'''


if __name__ == '__main__':
    setup(
        packages=find_packages(),
        name='CherryOnTop',
        version='0.0.3',
        author='Christopher Sira',
        author_email='cbsira@gmail.com',
        license='BSD',
        url='https://github.com/csira/cherryontop',
        description=('Helper abstractions and utilities for building JSON '
                     'APIs with CherryPy.'),
        long_description=long_desc,
        install_requires=[
            'cherrypy',
            'routes',
            'ujson',
        ],
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Environment :: Web Environment',
            'Framework :: CherryPy',
            'Intended Audience :: Developers',
            'License :: Freely Distributable',
            'License :: OSI Approved :: BSD License',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Topic :: Internet :: WWW/HTTP',
            'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
            'Topic :: Software Development :: Libraries :: Application Frameworks',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
    )



def route(uri_pattern, method='GET', methods=None):
    methods = methods or [method]
    new_routes = [(uri_pattern, method.upper(),) for method in methods]

    def wrap(f):
        routes = getattr(f, '_cbs_routes', [])
        routes += new_routes
        f._cbs_routes = routes
        return f

    return wrap


def delete(uri_pattern):
    return route(uri_pattern, method='DELETE')


def get(uri_pattern):
    return route(uri_pattern, method='GET')


def post(uri_pattern):
    return route(uri_pattern, method='POST')


def put(uri_pattern):
    return route(uri_pattern, method='PUT')
